@@ -4,6 +4,20 @@ Created on Thu Aug 10 11:23:50 2023
 
 @author: akomarla
 """
+import pandas as pd
+import numpy as np
+import logging
+import os
+import pyodbc as po
+from datetime import datetime
+from tqdm import tqdm
+import copy
+import matplotlib.pyplot as plt
+from statsmodels.tsa.holtwinters import SimpleExpSmoothing, Holt
+import sklearn.metrics as metrics
+from sklearn.model_selection import train_test_split
+from scipy import stats
+
 
 class TimeSeries:
     def __init__(self, values, labels):
@@ -113,38 +127,66 @@ class TimeSeries:
         else:
             return wna
     
-    def detect_outliers(self):
+    def detect_outliers(self, how):
         """
+        Parameters
+        ----------
+        how : str
+            specify whether to remove outliers using the IQR method: 'iqr' or plain percentiles: 'percentile'
 
         Returns
         -------
         outliers : list
-            outlier values generated using the percentile and IQR method
+            outlier values generated using the percentile or IQR method ONLY if there are more than two values in the input time-series
 
         """
+        # Get the time-series values
         values = np.sort(self.values)
-        # Quartile ranges
-        q1 = np.percentile(values, 25, method = 'midpoint') 
-        q3 = np.percentile(values, 75, method = 'midpoint')
-        # Inter quartile range
-        iqr = q3 - q1
-        # Limits
-        low_lim = q1 - 1.5 * iqr
-        up_lim = q3 + 1.5 * iqr
-        # Find outliers
-        outliers = []
-        for v in values:
-            if v < low_lim or v > up_lim:
-                outliers.append(v)
-        # Assign outlier values to instance
-        self.outliers = outliers
-        return outliers
-    
-    def remove_outliers(self, inplace = False):
+        
+        # Check outliers only if there are more than 2 values in the time-series
+        if len(values) > 2: 
+            # Use IQR method
+            if how == 'iqr':
+                # Quartile ranges
+                q1 = np.percentile(values, 25, method = 'midpoint') 
+                q3 = np.percentile(values, 75, method = 'midpoint')
+                # Inter quartile range
+                iqr = q3 - q1
+                # Limits
+                low_lim = q1 - 1.5 * iqr
+                up_lim = q3 + 1.5 * iqr
+                # Find outliers
+                outliers = []
+                for v in values:
+                    if v < low_lim or v > up_lim:
+                        outliers.append(v)
+                # Assign outlier values to instance
+                self.outliers = outliers
+                return outliers
+            # Use only percentiles
+            if how == 'percentile':
+                # Quartile ranges
+                q1 = np.percentile(values, 25, method = 'midpoint') 
+                q3 = np.percentile(values, 75, method = 'midpoint')
+                # Find outliers
+                outliers = []
+                for v in values:
+                    if v < q1 or v > q3:
+                        outliers.append(v)
+                # Assign outlier values to instance
+                self.outliers = outliers
+                return outliers
+        # If there are less than or equal to 2 values in the time-series then outliers will be NaN
+        else:
+            return [np.nan]
+        
+    def remove_outliers(self, how, inplace = False):
         """
 
         Parameters
         ----------
+        how : str
+            specify whether to remove outliers using the IQR method: 'iqr' or just percentiles: 'percentile'
         inplace : boolean, optional
             specify whether to replace the time-series values with modified values or not. The default is False.
 
@@ -155,7 +197,7 @@ class TimeSeries:
 
         """
         # Find the outlier values
-        _ = self.detect_outliers()
+        _ = self.detect_outliers(how = how)
         # Remove outliers from the time-series
         woo = [v for v in self.values if v not in self.outliers]
         # Replace time series values
@@ -207,8 +249,8 @@ class ExpSmoothing():
         ft = fit.forecast(num_gen)
         return ft
         
-    def train(self, train_data, train_true_values, error, num_gen, 
-              remove_outliers = False, non_neg = False, non_zero = False):
+    def train(self, train_data, train_true_values, num_gen = 1, error = 'mean absolute percentage error', 
+              remove_outliers = False, how = 'percentile', non_neg = False, non_zero = False):
         """
 
         Parameters
@@ -218,11 +260,13 @@ class ExpSmoothing():
         train_true_values : list of lists of floats
             true future values of the training data, ex: [[4,5], [7,8],...]
         error : string
-            type of error to use to train the model, ex: 'mean square error', 'mean absolute error'
+            type of error to use to train the model, ex: 'mean square error', 'mean absolute error'. Default is 'mean absolute error'
         num_gen : int
-            number of future values in the time-series to generate, ex: 1
+            number of future values in the time-series to generate, ex: 1. Default is 1.
         remove_outliers : boolean, optional
             remove outliers in the time-series to train the model. Default is False
+        how : str
+            specify whether to remove outliers using the IQR method: 'iqr' or just percentiles: 'percentile'. Default is 'percentile'
         non_neg : boolean, optional 
             remove negative values in the time-series to train the model. Default is False
         non_zero : boolean, optional 
@@ -250,13 +294,13 @@ class ExpSmoothing():
                     _ts = TimeSeries(values = ts, labels = None)
                     # Make modifications to the time series
                     _ts.clean(inplace = True, drop_na = True)
-                    # Remove outliers and non-negative values if specified 
-                    if remove_outliers:
-                        _ = _ts.remove_outliers(inplace = True)
+                    # Remove outliers and non-negative values if specified
                     if non_neg:
                         _ = _ts.non_neg(inplace = True)
                     if non_zero: 
                         _ = _ts.non_zero(inplace = True)
+                    if remove_outliers:
+                        _ = _ts.remove_outliers(how, inplace = True)
                     # Run the model
                     ft = self.es_model(ts = _ts.values, smoothing_level = alpha, num_gen = self.num_gen)
                     ft_val.append(ft.tolist())
@@ -298,11 +342,13 @@ class ExpSmoothing():
         self.param['alpha'] = min(e_dict, key = e_dict.get)
         # Store the chosen error type
         self.error = error
-        # Store the training error associated with the best alpha
+        # Store the training error associated with the best alpha values
         self.train_error = min(e_dict.values())
+        # Store all the training erros and alpha parameter values
+        self.train_log = e_dict
             
     def test(self, test_data, test_true_values, error = None, num_gen = None, param = None, 
-             remove_outliers = False, non_neg = False, non_zero = False):
+             remove_outliers = False, how = 'percentile', non_neg = False, non_zero = False):
         """
 
         Parameters
@@ -325,6 +371,9 @@ class ExpSmoothing():
             ignore to test the model using the optimal parameters from the training process
         remove_outliers : boolean, optional
             remove outliers in the time-series to test the model. Default is False
+        how : str
+            specify whether to remove outliers using the IQR method: 'iqr' or just percentiles: 'percentile'
+            default is 'percentile'
         non_neg : boolean, optional 
             remove negative values in the time-series to test the model. Default is False
         non_zero : boolean, optional 
@@ -360,12 +409,12 @@ class ExpSmoothing():
                 # Make modifications to the time series
                 _ts.clean(inplace = True, drop_na = True)
                 # Remove outliers and non-negative values if specified 
-                if remove_outliers:
-                    _ = _ts.remove_outliers(inplace = True)
                 if non_neg:
                     _ = _ts.non_neg(inplace = True)
                 if non_zero: 
                     _ = _ts.non_zero(inplace = True)
+                if remove_outliers:
+                    _ = _ts.remove_outliers(how = how, inplace = True)
                 # Run the model
                 ft = self.es_model(ts = _ts.values, smoothing_level = param['alpha'], num_gen = num_gen)
                 ft_val.append(ft.tolist())
@@ -404,7 +453,7 @@ class ExpSmoothing():
             
         return ft_val, self.test_error
     
-    def predict(self, data, num_gen = None, param = None, remove_outliers = False, non_neg = False, non_zero = False):
+    def predict(self, data, num_gen = None, param = None, remove_outliers = False, how = 'percentile', non_neg = False, non_zero = False):
         """
 
         Parameters
@@ -421,6 +470,9 @@ class ExpSmoothing():
             ignore to forecast values using the optimal parameters learned from model training process
         remove_outliers : boolean, optional
             remove outliers in the time-series. Default is False
+        how : str
+            specify whether to remove outliers using the IQR method: 'iqr' or just percentiles: 'percentile'
+            default is 'percentile'
         non_neg : boolean, optional 
             remove negative values in the time-series. Default is False
         non_zero : boolean, optional 
@@ -446,6 +498,13 @@ class ExpSmoothing():
                 _ts = TimeSeries(values = ts, labels = None)
                 # Make modifications to the time series
                 _ts.clean(inplace = True, drop_na = True)
+                # Remove outliers and non-negative values if specified 
+                if non_neg:
+                    _ = _ts.non_neg(inplace = True)
+                if non_zero: 
+                    _ = _ts.non_zero(inplace = True)
+                if remove_outliers:
+                    _ = _ts.remove_outliers(how = how, inplace = True)
                 ft = self.es_model(ts = _ts.values, smoothing_level = param['alpha'], num_gen = num_gen)
                 ft_val.append(ft.tolist())
             except:
@@ -499,7 +558,7 @@ class Analyze(TimeSeries, ExpSmoothing):
             except:
                 return np.nan
         
-    def forecast(self, how, num_gen = None, param = None):
+    def forecast(self, how, num_gen = None, param = None, remove_outliers = False, outliers_how = 'percentile', non_neg = False, non_zero = False):
         """
 
         Parameters
@@ -510,6 +569,15 @@ class Analyze(TimeSeries, ExpSmoothing):
             number of future values to generate using exponential smoothing forecast, ex: 1,2. Default is None
         param : dict, optional
             parameters needed for the exponential smoothing forecast, ex: {'alpha': 0.5}. Default is None
+        remove_outliers : boolean, optional
+            remove outliers in the time-series. Default is False
+        outliers_how : str
+            specify whether to remove outliers using the IQR method: 'iqr' or just percentiles: 'percentile'
+            default is 'percentile'
+        non_neg : boolean, optional 
+            remove negative values in the time-series. Default is False
+        non_zero : boolean, optional 
+            remove zero values in the time-series. Default is False
 
         Returns
         -------
@@ -528,7 +596,9 @@ class Analyze(TimeSeries, ExpSmoothing):
         elif how == 'average non-neg':
             return self.average(non_neg = True)
         elif how == 'exponential smoothing':
-            return self.predict(data = [self.values], num_gen = num_gen, param = param)
+            return self.predict(data = [self.values], num_gen = num_gen, param = param,
+                                remove_outliers = remove_outliers, how = outliers_how, 
+                                non_neg = non_neg, non_zero = non_zero)
 
 
 
@@ -868,7 +938,7 @@ class TimeInstance:
         fill : list
             year and quarter values, ex: ['2021Q4', '2021Q3', '2021Q2'....]
         """
-        return self.gen_quart_range(how = 'backward', num = num)[-1]
+        return self.gen_quart_range(how = 'backward', num = num)[0]
     
     def add_months(self, num, index = None):
         """
@@ -901,4 +971,4 @@ class TimeInstance:
         fill : list
             year and month values, ex: ['202104', '202103', '202102'....]
         """
-        return self.gen_month_range(how = 'backward', num = num, index = index)[-1]
\ No newline at end of file
+        return self.gen_month_range(how = 'backward', num = num, index = index)[0]
\ No newline at end of file
